{% extends "layout.jinja2" %}

{% block content %}
<div class="content">
  <h1>404 Not Found</h1>
    <p>Would you like to go <a href="/">back to the list of simulations?</a></p>
</div>
{% endblock content %}
