from .one_hop import HopStrategy, HopMajorDescendantStrategy, HopMajorProgenitorStrategy
from .multi_hop import MultiHopStrategy
from .multi_source import MultiSourceMultiHopStrategy
from .multi_hop_variants import MultiHopMajorDescendantsStrategy, MultiHopAllProgenitorsStrategy, \
    MultiHopMajorProgenitorsStrategy, MultiHopMostRecentMergerStrategy




