<!DOCTYPE html>
<html lang="{{request.locale_name}}">
  <head>
    <meta charset="utf-8">
    <!-- <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1.0"> -->
    <meta name="description" content="TANGOS: the amazing numerical galaxy organization system">
    <meta name="author" content="TANGOS collaboration">
      <link href="https://fonts.googleapis.com/css?family=Quicksand:400,500,700" rel="stylesheet">
    <link rel="shortcut icon" href="{{request.static_url('tangos.web:static/pyramid-16x16.png')}}">

    <title>TANGOS</title>


    <link href="{{request.static_url('tangos.web:static/stylesheet.css')}}" rel="stylesheet">
      <link href="{{request.static_url('tangos.web:static/jquery-ui.min.css')}}" rel="stylesheet">

    <!-- HTML5 shim and Respond.js IE8 support of HTML5 elements and media queries -->
    <!--[if lt IE 9]>
      <script src="//oss.maxcdn.com/libs/html5shiv/3.7.0/html5shiv.js" integrity="sha384-0s5Pv64cNZJieYFkXYOTId2HMA2Lfb6q2nAcx2n0RTLUnCAoTTsS0nKEO27XyKcY" crossorigin="anonymous"></script>
      <script src="//oss.maxcdn.com/libs/respond.js/1.3.0/respond.min.js" integrity="sha384-f1r2UzjsxZ9T4V1f2zBO/evUqSEOpeaUUZcMTz1Up63bl4ruYnFYeM+BxI4NhyI0" crossorigin="anonymous"></script>
    <![endif]-->

      <script type="text/javascript" src="{{ request.static_url('tangos.web:static/sorttable.js') }}"></script>
      <script type="text/javascript" src="{{ request.static_url('tangos.web:static/jquery-3.2.1.min.js') }}"></script>
      <script type="text/javascript" src="{{ request.static_url('tangos.web:static/jquery-ui.min.js') }}"></script>
      <script type="text/javascript" src="{{ request.static_url('tangos.web:static/autocomplete.js') }}"></script>
      <script type="text/javascript" src="{{ request.static_url('tangos.web:static/dragtable.js') }}"></script>
      <script type="text/javascript" src="{{ request.static_url('tangos.web:static/dynamictable.js') }}"></script>
      <script type="text/javascript" src="{{ request.static_url('tangos.web:static/keeponscreen.js') }}"></script>
      <script type="text/javascript" src="{{ request.static_url('tangos.web:static/pageupdate.js') }}"></script>
      <!-- NB formvalues.js must be included after dynamictable.js to make sure page elements are restored in correct order -->
      <script type="text/javascript" src="{{ request.static_url('tangos.web:static/formvalues.js') }}"></script>
      <script type="text/javascript" src="{{ request.static_url('tangos.web:static/mergertree.js') }}"></script>
      <script type="text/javascript" src="{{ request.static_url('tangos.web:static/d3.min.js') }}"></script>
  </head>

  <body>

    <div class="content" id="content-container">
      <div id="breadcrumbs-top" class="dynamic-update">{{ request.breadcrumbs | safe }}</div>
          {% block content %}
            <div>No content!</div>
        {% endblock content %}
      <div id="breadcrumbs-bottom" class="dynamic-update">{{ request.breadcrumbs | safe }}</div>
    </div>

    <!-- Bootstrap core JavaScript
    ================================================== -->
    </body>
</html>
